#include <stdio.h>
#include <stdlib.h>
#include <simu.h>

#define HASHPRIME 337
#define MYHASH(x) ((x) % HASHPRIME)

Hashnode **add_hash_tail (Hashnode ** tailaddr, int newfgl, int newindex)
/* adds a new node a linked list of hashnodes */
{
   Hashnode *out;
   out = malloc (sizeof (Hashnode));
   out->fgl = newfgl;
   out->index = newindex;
   out->next = NULL;
   *tailaddr = out;
   return &((*tailaddr)->next);
}

Hashnode **make_hashtable ()
/* the table is just an array of HASHPRIME pointers to hashnodes,
 * indexed by the hashes. fgls which have the same
 * hash are stored as a linked list */
{
   Hashnode **out;
   int   i;

   out = malloc (HASHPRIME * sizeof (Hashnode));
   for (i = 0; i < HASHPRIME; i++)
   {
      out[i] = NULL;
   }
   return out;
}

int get_index (int fgl, Hashnode ** hashtable, int *nextindex)
/* looks up an fgl in the hash table and adds a new 
 * node if it's not found; nextindex is the next available
 * input haplotype, so it gets assigned and incremented
 * if we add a new node */
{
   Hashnode **curr;

   int   b = MYHASH (fgl);
   curr = &hashtable[b];
   while (*curr)
   {
      if ((*curr)->fgl == fgl)
      {
         return (*curr)->index;
      }
      curr = &((*curr)->next);
   }
   /* if we get here, we didn't find the fgl */
   add_hash_tail (curr, fgl, (*nextindex)++);
   return (*curr)->index;
}

void print_table (FILE * outfile, Hashnode ** hashtable)
/* print out the fgl-to-column mappings stored in the hash table */
{
   int   i;
   Hashnode **curr;
   for (i = 0; i < HASHPRIME; i++)
   {
      curr = &hashtable[i];
      while (*curr)
      {
         fprintf (outfile, "%d %d\n", (*curr)->fgl, (*curr)->index);
         curr = &((*curr)->next);
      }
   }

}

Person read_person (FILE * fp)
    /* read in a Person struct--a bunch of id #s, gender, etc, then pointers
     * to chromosomes.  This function and read_chrom depend entirely on
     * the format of the file being read in, so be sure to update them 
     * together */
{
   Person out;

   fscanf (fp, "%d %*d %d %d %d %*d", &out.id, &out.motherid, &out.fatherid,
           &out.sex);
   out.maternal = read_chrom (fp);
   /* junk the id numbers on the next line--we already read them */
   fscanf (fp, "%*d %*d %*d %*d %*d %*d");
   out.paternal = read_chrom (fp);

   return out;
}

Chromseglist *read_chrom (FILE * fp)
/* read in junction data starting AFTER the id #s at start of the line */
{
   Chromseglist *out = NULL;
   Chromseglist **tailaddr;
   int   tempfgl;
   long long tempendpos;

   tailaddr = &out;

   /* kludge: fscanf zooms past newline characters, so we step back one
    * character after every read and check to see if the last 
    * skipped whitespace was a newline. in theory the line can be arbitrarily
    * large so we can't buffer an entire line. the "[]" has to be
    * in the fscanf pattern. fscanf contains deep mysteries. */
   do
   {
      fscanf (fp, "%d %lld []", &tempfgl, &tempendpos);
      tailaddr = add_chrom_tail (tailaddr, tempfgl, tempendpos);
      fseek (fp, -1, SEEK_CUR);
   }
   while (getc (fp) != '\n');
   return out;
}


/*reads file with snps(rows) and haplotypes(cols), no gaps.*/
char ** populate_snps (int nsnps, int npersons, FILE * fp)
{
   char **out;
   int   i, j;
   out = malloc (nsnps * sizeof (char *));
   for (i = 0; i < nsnps; i++)
   {
      out[i] = malloc (npersons * sizeof (char));
   }
   for (i = 0; i < nsnps; i++)
   {
      for (j = 0; j < npersons; j++)
      {
         out[i][j] = fgetc (fp);
      }
      fgetc (fp);
   }
   return out;
}

/*reads file with haplotypes(rows) and snps(cols), no gaps.*/
char ** populate_snps_new (int nsnps, int nhaps, FILE * fp)
{
   char **out;
   int   i, j;
   out = malloc (nsnps * sizeof (char *));
   for (i = 0; i < nsnps; i++)
   {
      out[i] = malloc (nhaps * sizeof (char));
   }
   for (i = 0; i < nhaps; i++)
   {
      for (j = 0; j < nsnps; j++)
      {
         out[j][i] = fgetc (fp);
      }
      fgetc (fp);
   }
   return out;
}

Chromseglist *get_ibd (Chromseglist * first, Chromseglist * second)
/* the algorithm here is roughly:
 * - set a pointer to the first node of each chromosome
 * - set last_end == 0 
 * 1. check for equality of fgl 
 * 1. a. if yes, we have entered an IBD segment
 *       keep advancing both pointers
 *      until they are different (meaning we
 *      have left the IBD segment)
 *    b. the IBD seg starts at "last_end" and ends
 *      at the lesser of the two pointers
 * 2. set "last_end" to the endpos of the most recently
 *    moved pointer (pointed to by curr_ptr)
 * 3. advance the lesser of the two pointers
 * 4. go to 1
 *
 * The idea is to zig-zag up the pair of chromosomes
 * from this:
 * ______*______________________
 *      /
 * ____/________________________
 *    *
 * to this:
 * ______*______________________
 *        \
 * ________\____________________
 *          *
 *
 * comparing fgl at each step
 *
 * Seems to work in basic cases but need to make sure
 * it's working at beginning or end of chromosome 
 */
{
   Chromseglist *out = NULL;
   Chromseglist **tail_ibd;
   Chromseglist *curr_first, *curr_second;
   Chromseglist **curr_ptr;
   Chromseglist *prev_ptr = NULL;
   long long last_end;

   tail_ibd = &out;
   curr_first = first;
   curr_second = second;
   curr_ptr = &curr_first;
   last_end = 0;

   while (*curr_ptr)
   {
      /* select smaller junction */
      if (curr_first->current->endpos < curr_second->current->endpos)
         curr_ptr = &curr_first;
      else
         curr_ptr = &curr_second;

      /* have we started an IBD segment? */
      if (curr_first->current->fgl == curr_second->current->fgl)
      {
         while (curr_first->next && curr_second->next &&
                curr_first->current->endpos == curr_second->current->endpos)
            /* advance through the IBD segment */
         {
            curr_first = curr_first->next;
            curr_second = curr_second->next;
         }

         /* IBD segment ends at smaller of first two unequal breaks */
         if (curr_first->current->endpos < curr_second->current->endpos)
            curr_ptr = &curr_first;
         else
            curr_ptr = &curr_second;

         tail_ibd = add_chrom_tail (tail_ibd, 0, last_end);
         tail_ibd = add_chrom_tail (tail_ibd, 1, (*curr_ptr)->current->endpos);
      }

      /* advance the smaller junction */
      last_end = (*curr_ptr)->current->endpos;
      prev_ptr = (*curr_ptr);
      (*curr_ptr) = (*curr_ptr)->next;
   }
   /* this is sloppy: tack on a no-ibd segment at the end,
    * with endpoint the end of the chromosome.
    * if there's already a yes-ibd segment there,
    * no matter, we'll never get there */
   tail_ibd = add_chrom_tail (tail_ibd, 0, prev_ptr->current->endpos);
   return out;
}

/* yanked from simu.c */
Chromseglist **add_chrom_tail (Chromseglist ** tailaddr, int newfgl,
                               long long newendpos)
{
   /* takes a Chromseglist pointer by reference, creates a new node,
    * and changes the passed pointer to point to the new node */
   Chromseglist *out;
   out = (Chromseglist *) malloc (sizeof (Chromseglist));
   out->current = (Chromseg *) malloc (sizeof (Chromseg));
   out->current->fgl = newfgl;
   out->current->endpos = newendpos;
   out->next = NULL;
   *tailaddr = out;
   return &((*tailaddr)->next);
}

int  *ibd_fifteen (Person a, Person b, int nmarks, int *snp_pos)
{
   /* we can identify the haplotype-based IBD status (15 possible states 
    * by testing each pair of haplotypes. each pattern of binary 
    * ibd states maps to an integer in [0,2^6-1].  we then map
    * that integer to the ibd state 1:15
    *
    * state pattern  integer
    *  1    1 1 1 1  63
    *  2    1 1 2 2  33
    *  3    1 1 1 2  11
    *  4    1 1 2 1  21
    *  5    1 1 2 3  1
    *  6    1 2 1 1  38
    *  7    1 2 2 2  56
    *  8    1 2 3 3  32
    *  9    1 2 1 2  18
    * 10    1 2 2 1  12
    * 11    1 2 1 3  2 
    * 12    1 2 3 1  4
    * 13    1 2 2 3  8
    * 14    1 2 3 2  16
    * 15    1 2 3 4  0
    *
    */
   int  *out;
   int   i, j;
   int   code;
   Chromseglist *tests[6];
   Chromseglist *curr[6];

   out = malloc (nmarks * sizeof (int));
   /* assemble Chromseglists with fgls either 1 or 0,
    * indicating pairwise IBD segments */
   tests[0] = get_ibd (a.maternal, a.paternal);
   tests[1] = get_ibd (a.maternal, b.maternal);
   tests[2] = get_ibd (a.maternal, b.paternal);
   tests[3] = get_ibd (a.paternal, b.maternal);
   tests[4] = get_ibd (a.paternal, b.paternal);
   tests[5] = get_ibd (b.maternal, b.paternal);

   for (j = 0; j < 6; j++)
   {
      curr[j] = tests[j];
   }

   for (i = 0; i < nmarks; i++)
   {
      code = 0;
      for (j = 0; j < 6; j++)
      {
         /* fast forward each IBD list */
         while (curr[j]->current->endpos < snp_pos[i])
         {
            curr[j] = curr[j]->next;
         }
         /* add test status times 2^j */
         code += curr[j]->current->fgl << j;
      }
      switch (code)
      {
          case 63:
             out[i] = 1;
             break;
          case 33:
             out[i] = 2;
             break;
          case 11:
             out[i] = 3;
             break;
          case 21:
             out[i] = 4;
             break;
          case 1:
             out[i] = 5;
             break;
          case 38:
             out[i] = 6;
             break;
          case 56:
             out[i] = 7;
             break;
          case 32:
             out[i] = 8;
             break;
          case 18:
             out[i] = 9;
             break;
          case 12:
             out[i] = 10;
             break;
          case 2:
             out[i] = 11;
             break;
          case 4:
             out[i] = 12;
             break;
          case 8:
             out[i] = 13;
             break;
          case 16:
             out[i] = 14;
             break;
          case 0:
             out[i] = 15;
             break;
          default:
             fprintf (stderr, "Unrecognized IBD pattern");
      }
   }
   return out;
}

int  *ibd_nine (int *ibd_fifteen, int nmarks)
{
   /* the genotype-based statuses are functions of the 15-state statuses */
   int   i;
   int  *out;
   out = malloc (nmarks * sizeof (int));
   for (i = 0; i < nmarks; i++)
   {
      switch (ibd_fifteen[i])
      {
          case 1:
             out[i] = 1;
             break;
          case 2:
             out[i] = 2;
             break;
          case 3:
          case 4:
             out[i] = 3;
             break;
          case 5:
             out[i] = 4;
             break;
          case 6:
          case 7:
             out[i] = 5;
             break;
          case 8:
             out[i] = 6;
             break;
          case 9:
          case 10:
             out[i] = 7;
             break;
          case 11:
          case 12:
          case 13:
          case 14:
             out[i] = 8;
             break;
          case 15:
             out[i] = 9;
             break;
      }
   }
   return out;
}
